from __future__ import division, print_function

from apache_beam import PTransform, ParDo, DoFn, Create
from kafka import KafkaConsumer, KafkaProducer


class KafkaConsume(PTransform):
    """A :class:`~apache_beam.transforms.ptransform.PTransform` for reading from an Apache Kafka topic. This is a streaming
    Transform that never returns. The transform uses `KafkaConsumer` from the
    `kafka` python library.

    It outputs a :class:`~apache_beam.pvalue.PCollection` of
    ``key-values:s``, each object is a Kafka message in the form (msg-key, msg)

    Args:
        consumer_config (dict): the kafka consumer configuration. The topic to
            be subscribed to should be specified with a key called 'topic'. The
            remaining configurations are those of `KafkaConsumer` from the
            `kafka` python library.
        value_decoder (function): Optional function to decode the consumed
            message value. If not specified, "bytes.decode" is used by default.
            "bytes.decode" which assumes "utf-8" encoding.

    Examples:
        Consuming from a Kafka Topic `notifications` ::

            import apache_beam as beam
            from apache_beam.options.pipeline_options import PipelineOptions
            from beam_nuggets.io import kafkaio

            consumer_config = {"topic": "notifications",
                               "bootstrap_servers": "localhost:9092",
                               "group_id": "notification_consumer_group"}

            with beam.Pipeline(options=PipelineOptions()) as p:
                notifications = p | "Reading messages from Kafka" >> kafkaio.KafkaConsume(
                    consumer_config=consumer_config,
                    value_decoder=bytes.decode,  # optional
                )
                notifications | 'Writing to stdout' >> beam.Map(print)

        The output will be something like ::

            ("device 1", {"status": "healthy"})
            ("job #2647", {"status": "failed"})

        Where the first element of the tuple is the Kafka message key and the second element is the Kafka message being passed through the topic
    """

    def __init__(self, consumer_config, value_decoder=None, *args, **kwargs):
        """Initializes ``KafkaConsume``
        """
        super(KafkaConsume, self).__init__()
        self._consumer_args = dict(
            consumer_config=consumer_config,
            value_decoder=value_decoder,
        )

    def expand(self, pcoll):
        return (
            pcoll
            | Create([self._consumer_args])
            | ParDo(_ConsumeKafkaTopic())
        )


class _ConsumeKafkaTopic(DoFn):
    """Internal ``DoFn`` to read from Kafka topic and return messages"""

    def process(self, consumer_args):
        consumer_config = consumer_args.pop('consumer_config')
        topic = consumer_config.pop('topic')
        value_decoder = consumer_args.pop('value_decoder') or bytes.decode
        consumer = KafkaConsumer(topic, **consumer_config)

        for msg in consumer:
            try:
                yield msg.key, value_decoder(msg.value)
            except Exception as e:
                print(e)
                continue


class KafkaProduce(PTransform):
    """A :class:`~apache_beam.transforms.ptransform.PTransform` for pushing messages
    into an Apache Kafka topic. This class expects a tuple with the first element being the message key
    and the second element being the message. The transform uses `KafkaProducer`
    from the `kafka` python library.

    Args:
        topic: Kafka topic to publish to
        servers: list of Kafka servers to listen to

    Examples:
        Examples:
        Pushing message to a Kafka Topic `notifications` ::

            from __future__ import print_function
            import apache_beam as beam
            from apache_beam.options.pipeline_options import PipelineOptions
            from beam_nuggets.io import kafkaio

            with beam.Pipeline(options=PipelineOptions()) as p:
                notifications = (p 
                                 | "Creating data" >> beam.Create([('dev_1', '{"device": "0001", status": "healthy"}')])
                                 | "Pushing messages to Kafka" >> kafkaio.KafkaProduce(
                                                                                        topic='notifications',
                                                                                        servers="localhost:9092"
                                                                                    )
                                )
                notifications | 'Writing to stdout' >> beam.Map(print)

        The output will be something like ::

            ("dev_1", '{"device": "0001", status": "healthy"}')

        Where the key is the Kafka topic published to and the element is the Kafka message produced
    """

    def __init__(self, topic=None, servers='127.0.0.1:9092'):
        """Initializes ``KafkaProduce``
        """
        super(KafkaProduce, self).__init__()
        self._attributes = dict(
            topic=topic, 
            servers=servers)

    def expand(self, pcoll):
        return (
            pcoll
            | ParDo(_ProduceKafkaMessage(self._attributes))
        )


class _ProduceKafkaMessage(DoFn):
    """Internal ``DoFn`` to publish message to Kafka topic"""

    def __init__(self, attributes, *args, **kwargs):
        super(_ProduceKafkaMessage, self).__init__(*args, **kwargs)
        self.attributes = attributes

    def start_bundle(self):
        self._producer = KafkaProducer(bootstrap_servers=self.attributes["servers"])

    def finish_bundle(self):
        self._producer.close()

    def process(self, element):
        try:
            self._producer.send(self.attributes['topic'], element[1].encode(), key=element[0].encode())
            yield element
        except Exception as e:
            raise



from __future__ import division, print_function

import apache_beam as beam


class AssignUniqueId(beam.DoFn):
    def __init__(self, id_key='id', *unused_args, **unused_kwargs):
        super(AssignUniqueId, self).__init__(*unused_args, **unused_kwargs)
        self.id_key = id_key
        self.id_counter = 0  # FIXME likely not enough for parallel jobs

    def process(self, element):
        """
        Args:
            element(dict):
        """
        element.update({self.id_key: self.id_counter})
        self.id_counter += 1
        yield element


from __future__ import division, print_function

import json

import apache_beam as beam
from beam_nuggets.compat import iteritems


class ParseJson(beam.DoFn):

    def __init__(self, only_keys=None, *unused_args, **unused_kwargs):
        self.only_keys = only_keys
        super(ParseJson, self).__init__(*unused_args, **unused_kwargs)

    def process(self, element):
        """
        Args:
            element(dict):
        """
        yield {
            k: json.loads(v) if k in self.only_keys else v
            for k, v in iteritems(element)
        }



from __future__ import division, print_function

import unittest

from apache_beam.testing.test_pipeline import TestPipeline
from apache_beam.testing.util import assert_that, equal_to

from beam_nuggets.io import relational_db
from .test_base import TransformBaseTest


class TestReadTransform(TransformBaseTest):

    def setUp(self):
        super(TestReadTransform, self).setUp()
        self.table_name, self.table_rows = (
            self.create_and_populate_test_table(n_rows=10)
        )

    def test_ReadFromRelationalDB(self):
        # create read pipeline, execute it and compare retrieved to actual rows
        with TestPipeline() as p:
            assert_that(
                p | "Reading records from db" >> relational_db.ReadFromDB(
                    source_config=self.source_config,
                    table_name=self.table_name
                ),
                equal_to(self.table_rows)
            )

    def test_QueryFromRelationalDB(self):
        # create a read pipeline with a SQL query, execute it and compare retrieved to actual rows
        with TestPipeline() as p:
            assert_that(
                p | "Reading records from db" >> relational_db.ReadFromDB(
                    source_config=self.source_config,
                    table_name=self.table_name,
                    query='select * from ' + self.table_name
                ),
                equal_to(self.table_rows)
            )

    def create_and_populate_test_table(self, n_rows=10):
        # test table schema and data
        table_name = 'students'
        ID, NAME, AGE = 'id', 'name', 'age'

        def define_table(metadata):
            from sqlalchemy import Table, Integer, String, Column
            return Table(
                table_name, metadata,
                Column(ID, Integer, primary_key=True),
                Column(NAME, String(100)),
                Column(AGE, Integer)
            )

        rows = [
            {ID: row_id, NAME: 'Jack{}'.format(row_id), AGE: 20 + row_id}
            for row_id in range(n_rows)
        ]

        # create test table
        self.db.create_table(
            name=table_name,
            define_table_f=define_table,
            create_table_if_missing=True
        )

        # populate
        self.db.write_rows(table_name, rows)
        return table_name, rows


if __name__ == '__main__':
    unittest.main()

